@@ -7,7 +7,6 @@ use core::{
     future::Future,
     pin::Pin,
     task::{Context, Poll},
-    time::Duration,
 };
 use futures::future::{Fuse, FutureExt};
 use pin_project::pin_project;
@@ -16,37 +15,15 @@ use s2n_quic_core::{
     event::{self, EndpointPublisher as _},
     inet::{self, SocketAddress},
     path::MaxMtu,
-    time::{self, Clock as ClockTrait},
+    time::Clock as ClockTrait,
 };
 use std::{convert::TryInto, io, io::ErrorKind};
-use tokio::{net::UdpSocket, runtime::Handle, time::Instant};
+use tokio::{net::UdpSocket, runtime::Handle};
 
 pub type PathHandle = socket::Handle;
 
-#[derive(Debug)]
-struct Clock(Instant);
-
-impl Default for Clock {
-    fn default() -> Self {
-        Self::new()
-    }
-}
-
-impl Clock {
-    pub fn new() -> Self {
-        Self(Instant::now())
-    }
-}
-
-impl ClockTrait for Clock {
-    fn get_time(&self) -> time::Timestamp {
-        let duration = self.0.elapsed();
-        unsafe {
-            // Safety: time duration is only derived from a single `Instant`
-            time::Timestamp::from_duration(duration)
-        }
-    }
-}
+mod clock;
+use clock::Clock;
 
 impl crate::socket::std::Socket for UdpSocket {
     type Error = io::Error;
@@ -513,12 +490,7 @@ impl<E: Endpoint<PathHandle = PathHandle>> Instance<E> {
             }
         }
 
-        /// Even if there is no progress to be made, wake up the task at least once a second
-        const DEFAULT_TIMEOUT: Duration = Duration::from_secs(1);
-
-        let mut prev_time = Instant::now() + DEFAULT_TIMEOUT;
-        let sleep = tokio::time::sleep_until(prev_time);
-        tokio::pin!(sleep);
+        let mut timer = clock.timer();
 
         loop {
             // Poll for readability if we have free slots available
@@ -545,70 +517,52 @@ impl<E: Endpoint<PathHandle = PathHandle>> Instance<E> {
             // pin the wakeups future so we don't have to move it into the Select future.
             tokio::pin!(wakeups);
 
-            let select = Select::new(rx_task, tx_task, &mut wakeups, &mut sleep);
-
-            let mut reset_clock = false;
-
-            if let Ok((rx_result, tx_result, timeout_result)) = select.await {
-                let subscriber = endpoint.subscriber();
-                let mut publisher = event::EndpointPublisherSubscriber::new(
-                    event::builder::EndpointMeta {
-                        endpoint_type: E::ENDPOINT_TYPE,
-                        timestamp: clock.get_time(),
-                    },
-                    None,
-                    subscriber,
-                );
-
-                publisher.on_platform_event_loop_wakeup(event::builder::PlatformEventLoopWakeup {
-                    timeout_expired: timeout_result,
-                    rx_ready: rx_result.is_some(),
-                    tx_ready: tx_result.is_some(),
-                });
-
-                if let Some(guard) = tx_result {
-                    if let Ok(result) = guard?.try_io(|socket| tx.tx(socket, &mut publisher)) {
-                        result?;
-                    }
-                }
-
-                // if the timer expired ensure it is reset to the default timeout
-                if timeout_result {
-                    reset_clock = true;
-                }
-
-                if let Some(guard) = rx_result {
-                    if let Ok(result) = guard?.try_io(|socket| rx.rx(socket, &mut publisher)) {
-                        result?;
-                    }
-                    endpoint.receive(&mut rx.rx_queue(), &clock);
-                }
+            let SelectOutcome {
+                rx_result,
+                tx_result,
+                timeout_expired,
+                application_wakeup,
+            } = if let Ok(res) = Select::new(rx_task, tx_task, &mut wakeups, &mut timer).await {
+                res
             } else {
                 // The endpoint has shut down
                 return Ok(());
+            };
+
+            let subscriber = endpoint.subscriber();
+            let mut publisher = event::EndpointPublisherSubscriber::new(
+                event::builder::EndpointMeta {
+                    endpoint_type: E::ENDPOINT_TYPE,
+                    timestamp: clock.get_time(),
+                },
+                None,
+                subscriber,
+            );
+
+            publisher.on_platform_event_loop_wakeup(event::builder::PlatformEventLoopWakeup {
+                timeout_expired,
+                rx_ready: rx_result.is_some(),
+                tx_ready: tx_result.is_some(),
+                application_wakeup,
+            });
+
+            if let Some(guard) = tx_result {
+                if let Ok(result) = guard?.try_io(|socket| tx.tx(socket, &mut publisher)) {
+                    result?;
+                }
+            }
+
+            if let Some(guard) = rx_result {
+                if let Ok(result) = guard?.try_io(|socket| rx.rx(socket, &mut publisher)) {
+                    result?;
+                }
+                endpoint.receive(&mut rx.rx_queue(), &clock);
             }
 
             endpoint.transmit(&mut tx.tx_queue(), &clock);
 
             if let Some(delay) = endpoint.timeout() {
-                let delay = unsafe {
-                    // Safety: the same clock epoch is being used
-                    delay.as_duration()
-                };
-
-                // floor the delay to milliseconds to reduce timer churn
-                let delay = Duration::from_millis(delay.as_millis() as u64);
-
-                // add the delay to the clock's epoch
-                let next_time = clock.0 + delay;
-
-                // if the clock has changed let the sleep future know
-                if next_time != prev_time {
-                    sleep.as_mut().reset(next_time);
-                    prev_time = next_time;
-                }
-            } else if reset_clock {
-                sleep.as_mut().reset(Instant::now() + DEFAULT_TIMEOUT);
+                timer.update(delay);
             }
         }
     }
@@ -659,7 +613,14 @@ where
     }
 }
 
-type SelectResult<Rx, Tx> = Result<(Option<Rx>, Option<Tx>, bool), CloseError>;
+struct SelectOutcome<Rx, Tx> {
+    rx_result: Option<Rx>,
+    tx_result: Option<Tx>,
+    timeout_expired: bool,
+    application_wakeup: bool,
+}
+
+type SelectResult<Rx, Tx> = Result<SelectOutcome<Rx, Tx>, CloseError>;
 
 impl<Rx, Tx, Wakeup, Sleep> Future for Select<Rx, Tx, Wakeup, Sleep>
 where
@@ -674,9 +635,11 @@ where
         let this = self.project();
 
         let mut should_wake = false;
+        let mut application_wakeup = false;
 
         if let Poll::Ready(wakeup) = this.wakeup.poll(cx) {
             should_wake = true;
+            application_wakeup = true;
             if let Err(err) = wakeup {
                 return Poll::Ready(Err(err));
             }
@@ -692,10 +655,10 @@ where
             *this.tx_out = Some(v);
         }
 
-        let mut timeout_result = false;
+        let mut timeout_expired = false;
 
         if this.sleep.poll(cx).is_ready() {
-            timeout_result = true;
+            timeout_expired = true;
             should_wake = true;
         }
 
@@ -704,7 +667,12 @@ where
             return Poll::Pending;
         }
 
-        Poll::Ready(Ok((this.rx_out.take(), this.tx_out.take(), timeout_result)))
+        Poll::Ready(Ok(SelectOutcome {
+            rx_result: this.rx_out.take(),
+            tx_result: this.tx_out.take(),
+            timeout_expired,
+            application_wakeup,
+        }))
     }
 }
 
@@ -766,7 +734,7 @@ mod tests {
             tx,
         },
         path::Handle as _,
-        time::{Clock, Timestamp},
+        time::{Clock, Duration, Timestamp},
     };
     use std::collections::BTreeMap;
 
